package tnntpd;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.util.LinkedList;
import java.util.UUID;

public class Article {

  private static final String NEWLINE = "\r\n" ;
  private static final String MESSAGE_ID_PATTERN = "<[^>]+>" ; 

  public static final String BODY_TYPE_TEXT = "text" ; 
  public static final String BODY_TYPE_FILE = "file" ; 
  public static final String BODY_TYPE_FILE_GZIP = "file.gz" ; 

  private int id ;
  private String messageId ;
  private Timestamp date ;
  
  private String body ;
  private String bodyType = BODY_TYPE_TEXT;  
  
  public static Article getByMessageId ( String id ) {
    if ( ! id.matches(MESSAGE_ID_PATTERN) ) return null;
    try {
      PreparedStatement stmt = NNTPConnection.getDB().prepareStatement("SELECT * FROM articles WHERE message_id = ?;") ;
      stmt.setString(1,id) ;
      return loadFromStatement(stmt) ;
    } catch ( SQLException e ) {
      return null ;
    }
  }
  
  public static Article getById ( int id ) {
    try {
      PreparedStatement stmt = NNTPConnection.getDB().prepareStatement("SELECT * FROM articles WHERE id = ?;") ;
      stmt.setInt(1,id) ;
      return loadFromStatement(stmt) ;
    } catch ( SQLException e ) {
      return null ;
    }
  }

  public static Article loadFromStatement ( PreparedStatement stmt ) {
    try {
      ResultSet rs = stmt.executeQuery() ;
      if ( ! rs.next() ) return null ;
      return loadFromResultSet(rs) ;    
    } catch ( SQLException e ) {
      return null ;
    }
  }

  public static Article loadFromResultSet ( ResultSet rs ) throws SQLException {
    if ( rs.isAfterLast() || rs.getInt("id") == 0 ) return null ;
    Article a = new Article () ;
    a.setId ( rs.getInt("id") ) ;
    a.setMessageId ( rs.getString("message_id") ) ;
    a.setDate ( rs.getTimestamp("date") ) ;
    a.setBody ( rs.getString("body") ) ;
    a.setBodyType( rs.getString("body_type") ) ;
    return a ;
  }
  
  public void create () throws SQLException {
    PreparedStatement stmt = NNTPConnection.getDB().prepareStatement("INSERT INTO articles (message_id,date,body,body_type) VALUES (?, ?, ?, ?);") ;
    stmt.setString(1,getMessageId()) ;
    stmt.setTimestamp(2,getDate()) ;
    stmt.setString(3,getBody()) ;
    stmt.setString(4,getBodyType()) ;
    stmt.execute() ;
    setId ( getByMessageId(getMessageId()).getId());
  }
    
  public Article nextArticleInGroup ( Group group ) {
    try {
      PreparedStatement stmt = NNTPConnection.getDB ().prepareStatement ("SELECT articles.* FROM articles, article_groups as ag WHERE articles.id = ag.article_id AND ag.group_id = ? AND article_id > ? ORDER BY articles.id ASC LIMIT 1" ) ;
      stmt.setInt ( 1, group.getId() ) ;
      stmt.setInt ( 2, getId() ) ;
      ResultSet rs = stmt.executeQuery () ;
      rs.next () ;
      return Article.loadFromResultSet ( rs ) ;
    } catch ( SQLException e ) {
      return null ;
    }
  }
  
  public Article prevArticleInGroup ( Group group ) {
    try {
      PreparedStatement stmt = NNTPConnection.getDB ().prepareStatement ("SELECT articles.* FROM articles, article_groups as ag WHERE articles.id = ag.article_id AND ag.group_id = ? AND article_id < ? ORDER BY articles.id DESC LIMIT 1" ) ;
      stmt.setInt ( 1, group.getId() ) ;
      stmt.setInt ( 2, getId() ) ;
      ResultSet rs = stmt.executeQuery () ;
      rs.next () ;
      return Article.loadFromResultSet ( rs ) ;
    } catch ( SQLException e ) {
      return null ;
    }
  }

  public void generateMessageId () {
    setMessageId ( "<" + UUID.randomUUID() + "@" + NNTPServer.hostname + ">" );
  }
  
  public LinkedList<Group> getGroups () throws SQLException {
    PreparedStatement stmt = NNTPConnection.getDB().prepareStatement("SELECT * FROM article_groups, groups WHERE article_id = ? AND group_id = groups.id ORDER BY name ASC;") ;
    stmt.setInt(1,getId()) ;
    LinkedList<Group> list = Group.loadAllFromStatement(stmt) ;
    return list ;
  }
  
  public OrderPreservingMap getHeaders () throws SQLException {
    PreparedStatement stmt = NNTPConnection.getDB().prepareStatement("SELECT * FROM headers WHERE article_id = ?;") ;
    stmt.setInt(1,getId()) ;
    OrderPreservingMap map = Header.loadAllFromStatement(stmt) ;
    return map ;
  }

  public void delete () throws SQLException {
    PreparedStatement stmt = NNTPConnection.getDB().prepareStatement("DELETE FROM articles WHERE id = ?") ;
    stmt.setInt(1,getId()) ;
    stmt.execute() ;
  }

  public CharSequence getXref () {
    StringBuilder sb = new StringBuilder () ;
    sb.append("Xref: " + NNTPServer.hostname ) ;
    try {
      for ( Group g : getGroups() ) {
        sb.append(" " + g.getName() + ":" + getId() ) ;
      }
    } catch ( SQLException e ) {}
    return sb ;
  }
  
  public void addHeader ( String name, String value ) throws SQLException {
    Header h = new Header () ;
    h.setArticleId(getId()) ;
    h.setName(name) ;
    h.setValue(value) ;
    h.create() ;
  }
  
  /* **************************************************************************
   *                         Getters and Setters
   ****************************************************************************/
  
  public String getBody () {
    return body ;
  }

  public void setBody ( String body ) {
    this.body = body ;
  }

  public Timestamp getDate () {
    return date ;
  }

  public void setDate ( Timestamp date ) {
    this.date = date ;
  }

  public int getId () {
    return id ;
  }

  public void setId ( int id ) {
    this.id = id ;
  }

  public String getMessageId () {
    return messageId ;
  }

  public void setMessageId ( String messageId ) {
    this.messageId = messageId ;
  }
  
  public String getBodyType () {
    return bodyType ;
  }

  public void setBodyType ( String bodyType ) {
    this.bodyType = bodyType ;
  }

  public CharSequence getCompleteHeader () {
    StringBuilder sb = new StringBuilder () ;
    try {
      OrderPreservingMap headers = getHeaders() ;
      for ( String name : headers ) sb.append ( name + ": " + headers.get(name) + NEWLINE ) ;
    } catch ( SQLException e ) {}
    sb.append(getXref() + NEWLINE) ;
    return sb ;
  }
    
}
