package tnntpd;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.Socket;
import java.net.SocketException;
import java.sql.Connection;
import java.sql.SQLException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.LinkedList;

import tnntpd.command.*;

public class NNTPConnection extends Thread {

  public static final String NEWLINE = "\r\n" ;
  public static final String MESSAGE_ID_PATTERN = "<[^>]+>" ; 

  private boolean debug = false ;
  
  private Socket socket ;
  private boolean exit = false ;
  private BufferedWriter out ;
  private BufferedReader in ;
  private Connection tempDB ;
  private static ThreadLocal<Connection> db = new ThreadLocal<Connection>();
  private static ThreadLocal<NNTPConnection> instance = new ThreadLocal<NNTPConnection>();
  private File articleDir = new File ( Main.settings.getProperty("tnntpd.articledir")) ;
  
  private Article currentArticle = null ;
  private Group currentGroup = null ;
  
  public NNTPConnection ( Socket socket, Connection db ) throws IOException {
    this.socket = socket;
    in = new BufferedReader ( new InputStreamReader ( socket.getInputStream() ), 1024 * 10 ) ;
    out = new BufferedWriter ( new OutputStreamWriter ( socket.getOutputStream() ) ) ;
    tempDB = db ;
    if ( Main.settings.getProperty("tnntpd.debug") != null ) debug = true ;
  }

  public static Connection getDB () {
    return db.get() ;
  }
  
  public static NNTPConnection getInstance () {
    return instance.get() ;
  }

  public void exit () {
    try {
      exit = true ;
      socket.close() ;
      getDB().close() ;
    } catch ( Exception e ) {}
  }
  
  /* ****************************************************************************
   *                            Printing to the stream 
   ******************************************************************************/
  
  public void print ( CharSequence s ) throws IOException {
    out.append(s) ;
  }
  
  public void println ( CharSequence s) throws IOException {
    print ( s ) ;
    print ( NEWLINE ) ;
    if ( debug ) System.out.println ( "<<< " + s ) ; 
  }
  
  public void printStatus (int code, String description) throws IOException {
    println ( "" + code + " " + description ) ;
    flush();
  }
  
  public void printTextLine (CharSequence line) throws IOException {
    if ( line.length() > 0 && line.charAt(0) == '.' ) print ( ".." ) ;
    println (line) ;    
  }
  
  public void printTextPart (BufferedReader text) throws IOException {
    String s ;
    do {
      s = text.readLine() ;
      if ( s == null ) return ;
      printTextLine ( s ) ;
    } while ( true ) ;
  }

  public void printTextPart (CharSequence text) throws IOException {
    String[] lines = text.toString().split(NEWLINE) ;
    for ( String line : lines ) printTextLine ( line ) ;
  }

  public void printText (BufferedReader text) throws IOException {
    printTextPart ( text ) ;
    println (".") ;
    flush();
  }

  public void printText (CharSequence text) throws IOException {
    printTextPart ( text ) ;
    println (".") ;
    flush();
  }
  
  public void flush () throws IOException {
    out.flush() ;
  }
  
  /* ****************************************************************************
   *                            Reading from the stream 
   ******************************************************************************/
  
  public String readln () throws IOException {
    String s = in.readLine() ;
    if ( s == null ) throw new IOException ( "Socket closed" ) ;
    if ( debug ) System.out.println ( ">>> " + s ) ;
    return s ;
  }
  
  public String[] readCommand () throws IOException {
    return readln().split("[ ]+") ;
  }
  
  public LinkedList<String> readText () throws IOException {
    LinkedList<String> l = new LinkedList<String> () ;
    String s ;
    do {
      s = readln() ;
      if ( ! s.equals(".") ) {
        if ( s.startsWith("..") ) s = s.substring(1) ;
        l.addLast(s);
      }
    } while ( ! s.equals(".") ) ;
    return l ;
  }
  
  public String readTextLine () throws IOException {
    String s = null ;
    do {
      s = readln() ;
    } while ( s == null ) ;
    if ( s.equals(".") ) return null ;
    if ( s.startsWith("..") ) s = s.substring(1) ;
    return s ;
  }

  /* ****************************************************************************
   *                            Article commands 
   ******************************************************************************/
  
  public File getArticleDir () {
    if ( ! articleDir.exists() ) articleDir.mkdir() ;
    return articleDir ;
  }
  
  public void setCurrentArticle ( Article current ) {
    currentArticle = current ;
  }
  
  public Article getCurrentArticle () {
    return currentArticle ;
  }
  
  /* ****************************************************************************
   *                            Group commands 
   ******************************************************************************/


  public void setCurrentGroup ( Group current ) {
    currentGroup = current ;
  }
  
  public Group getCurrentGroup () {
    return currentGroup ;
  }
  
  /* ****************************************************************************
   *                            Command processing
   ******************************************************************************/
  
  private void processCommand ( String[] command ) throws IOException {
    if ( command.length == 0 ) return ; // TODO Error
    String commandName = command[0] ;
    
    // RFC977
    // TODO HELP command
    // TODO NEWGROUPS command
    // TODO NEWNEWS command
    
    // RFC2980
    // TODO LIST ACTIVE command
    // TODO LIST ACTIVE.TIMES command
    // TODO LIST DISTRIBUTIONS command
    // TODO LIST DISTRIB.PATS command
    // TODO XGTITLE command
    // TODO XHDR command
    // TODO XPAT command
    // TODO XPATH command
    // TODO XROVER command
    // TODO XTHREAD command
    // TODO AUTHINFO command
    
    if ( ArticleCommand.process(command) || ListCommand.process(command)
      || NextCommand.process(command) || PostCommand.process(command) 
      || XoverCommand.process(command) || ListGroupCommand.process(command)
      || GroupCommand.process(command) || NewGroupsCommand.process(command) ) {
     
    }
    
    else if ( commandName.equalsIgnoreCase("CHECK") || commandName.equalsIgnoreCase("TAKETHIS") ) {
      // untested, RFC2980 complient
      printStatus (400, "not accepting articles" ) ;
      return;
    }
    
    else if ( commandName.equalsIgnoreCase("IHAVE") || commandName.equalsIgnoreCase("XREPLIC" ) ) {
      // untested, RFC977 complient
      printStatus (435, "article not wanted - do not send it" ) ;
      return;
    }
    
    else if ( commandName.equalsIgnoreCase("XCREATEGROUP") ) {
      // untested
      try {
        Group g = new Group () ;
        g.setName(command[1]) ;
        g.create() ;
        printStatus (280, "group created") ;
      } catch ( SQLException e ) {
        try {
          getDB().rollback() ;
        } catch ( SQLException e1 ) {}
        e.printStackTrace() ;
        printStatus (480, "group not created - database error") ;
      }
      return;
    }

    else if ( commandName.equalsIgnoreCase("SLAVE") ) {
      // untested, RFC977 complient
      printStatus (202, "slave status noted") ;
      return;
    }
    
    else if ( commandName.equalsIgnoreCase("XINDEX") ) {
      // untested, RFC2980 complient
      printStatus (418, "no tin-style index is available for this news group") ;
      return;
    }

    else if ( commandName.equalsIgnoreCase("DATE") ) {
      printStatus (111, new SimpleDateFormat ("yyyyMMddHHmmss").format(new Date()) ) ;
      return ;
    }
    
    else if ( commandName.equalsIgnoreCase("MODE") ) {
      if ( command[1].equalsIgnoreCase("READER") ) {
        // untested, RFC2980 complient
        printStatus (200, "Hello, you can post") ;
      } 
      else if ( command[1].equalsIgnoreCase("STREAM") ) {
        printStatus (203, "Streaming is OK") ;
      }
      else printStatus (501, "Command not supported") ;
    }

    else if ( commandName.equalsIgnoreCase("XSHUTDOWN") ) {
      printStatus (205, "closing connection - goodbye!") ;
      exit () ;
      NNTPServer.instance.exit() ;
      return;
    }

    else if ( commandName.equalsIgnoreCase("QUIT") ) {
      // untested, RFC977 complient
      printStatus (205, "closing connection - goodbye!") ;
      exit() ;
      return;
    }
    
    else printStatus (501, "Command not supported") ;
  }
  
  public void run () {
    if ( ! this.equals(Thread.currentThread()) ) throw new RuntimeException ( "Do not call run() directly !" ) ;
    db.set(tempDB);
    instance.set(this);
    try {
      printStatus (200, NNTPServer.hostname + " tnntpd news server ready - (posting ok).") ;
    } catch ( IOException e1 ) {
      exit() ;
    }
    while ( ! exit ) {
      try {
        processCommand ( readCommand() ) ;
      } catch ( SocketException e ) {
        if ( exit ) return ;
        exit() ;
        e.printStackTrace() ;
      } catch ( IOException e ) {
        if ( exit ) return ;
        exit();
        e.printStackTrace() ;
      } catch ( Throwable e ) {
        if ( exit ) return ;
        e.printStackTrace() ;
        // silently ignore
      }
    }
  }
  
}
