package tnntpd.command;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import java.util.UUID;
import java.util.zip.GZIPOutputStream;

import tnntpd.Article;
import tnntpd.Group;
import tnntpd.OrderPreservingMap;

public class PostCommand extends Command {

  public static int MAX_BODY_SIZE_IN_DB = 1024 ;
  public static int MAX_HEADER_SIZE = 10 * 1024 ;
  
  private static void deleteArticle ( Article article ) throws SQLException {
    if ( article.getBodyType().equalsIgnoreCase(Article.BODY_TYPE_FILE) 
      || article.getBodyType().equalsIgnoreCase(Article.BODY_TYPE_FILE_GZIP) ) 
      new File ( getArticleDir(), article.getBody() ).delete() ;
    article.delete() ;
  }
  
  public static boolean process ( String[] command ) throws IOException {
    if ( ! command[0].equalsIgnoreCase("POST") ) return false ;
    // untested, RFC977 complient
    
    printStatus (340, "send article to be posted. End with <CR-LF>.<CR-LF>") ;
    
    // some initialization
    Article article = new Article () ;
    int lineCount = 0;
    long bodySize = 0 ;

    // begin with a stringbuilder body 
    StringBuilder body = new StringBuilder () ;
    article.setBodyType(Article.BODY_TYPE_TEXT) ;
    Writer bodyFileWriter = null ;
    File bodyFile = null ;
    
    OrderPreservingMap headers = new OrderPreservingMap () ;
    
    boolean isHeader = true ; // are we in the header part

    String line = readTextLine() ;
    while ( line != null) {
      if ( ! isHeader ) { // body
        bodySize += line.length() + 1 ;
        lineCount++ ;
        if ( bodyFileWriter != null ) { // we are writing to a file
          bodyFileWriter.append(line+NEWLINE);
        } else { // we are still writing in the StringBuilder
          body.append(line+NEWLINE) ;
          if ( bodySize > MAX_BODY_SIZE_IN_DB ) { // too large for our StringBuilder
            // create the target file and setup the writer
            bodyFile = new File ( getArticleDir(), UUID.randomUUID().toString() ) ;
            bodyFile.createNewFile() ;
            bodyFileWriter = new OutputStreamWriter ( new GZIPOutputStream ( new FileOutputStream (bodyFile) ) ) ;
            // append the old body contents to the writer
            bodyFileWriter.append(body) ;
            // set the body type to file and store the filename in the body
            body = new StringBuilder() ;
            body.append( bodyFile.getName() );
            article.setBodyType(Article.BODY_TYPE_FILE_GZIP) ;            
          }
        }
      }
      if ( line.equals("") ) isHeader = false ; // we finally met the blank line separating headers from body
      if ( isHeader ) { // header
        // split name and value and add the header to the map
        int colon = line.indexOf(':') ;
        String fieldName = line.substring(0,colon) ;
        String fieldValue = line.substring(colon+1).trim() ;
        headers.append(fieldName,fieldValue) ;
      }
      line = readTextLine () ; // read a new line
    }
    if ( bodyFileWriter != null ) bodyFileWriter.close() ; // close the body file

    article.setBody(body.toString()) ; // set the article body
    
    // read the date header and fall back to the current date if it is not set
    try {
      article.setDate(new Timestamp(new SimpleDateFormat ("EEE, dd MMM yyyy HH:mm:ss Z",Locale.US).parse(headers.get("DATE")).getTime())) ;
    } catch ( NullPointerException e ) {
      article.setDate(new Timestamp(new Date().getTime())) ;
      headers.append("Date",new SimpleDateFormat ("EEE, dd MMM yyyy HH:mm:ss Z",Locale.US).format(article.getDate())) ;
    } catch ( ParseException e ) {
      printStatus ( 541, "posting failed - invalid date format" ) ;
      return true;
    }
        
    // read the message-id header
    if ( headers.containsKey("MESSAGE-ID") ) article.setMessageId(headers.get("MESSAGE-ID")) ;
    
    // check for a cancel command
    if ( headers.containsKey("CONTROL") ) {
      String[] control = headers.get("CONTROL").split(" ") ;
      if ( control.length >= 2 && control[0].equalsIgnoreCase("cancel") ) {
        // this article is a cancel-article, try to delete the old article
        try {
          deleteArticle ( Article.getByMessageId(control[1]) ) ;
          printStatus (240, "article posted ok - original article canceled") ; // quite confusing message 
          return true; // quit, do not actually post this article since it contains nonsense
        } catch ( SQLException e ) {
          // try a rollback and quit, do not actually post this article since it contains nonsense
          try {
            getDB().rollback() ;
          } catch ( SQLException e1 ) {}
          printStatus ( 441, "posting failed - original posting not found");
          return true;
        }
      }
    }
    
    // check for a supersede command
    if ( headers.containsKey("SUPERSEDES") ) {
      // just set set the message id to that of the original article
      article.setMessageId(headers.get("SUPERSEDES")) ;
      // try to delete the old article
      try {
        deleteArticle ( Article.getByMessageId(headers.get("SUPERSEDES")) ) ;
      } catch ( SQLException e ) {
        try {
          getDB().rollback() ;
        } catch ( SQLException e1 ) {}
        printStatus ( 441, "posting failed - original posting not found");
        return true; // quit, no article to supersede
      }
    }
    
    // if no message-id has been set so far, generate one
    if ( article.getMessageId() == null ) article.generateMessageId () ;
    
    // set some headers
    headers.set("Message-ID",article.getMessageId()) ;
    headers.set("Lines",""+lineCount) ;
    headers.set("Bytes",""+bodySize) ;

    // if needed, set an empty references header
    if ( ! headers.containsKey("REFERENCES") ) headers.set("References","");
        
    try {
      // try to create the article in the database
      article.create() ;
      // add the article to its groups
      String[] groupNames = headers.get("NEWSGROUPS").split(",") ;
      for ( String groupName: groupNames ) {
        Group g = Group.getByName(groupName.trim()) ;
        if ( g != null ) g.addArticle(article) ;
      }
      // add the header entries for the article
      for ( String headerName : headers ) {
        article.addHeader(headerName,headers.get(headerName)) ;
      }
      getDB().commit() ;
    } catch ( SQLException e ) {
      // if anything fails, rollback the whole thing
      try {
        getDB().rollback() ;
      } catch ( SQLException e1 ) {}
      e.printStackTrace() ;
      printStatus ( 441, "posting failed - database error");
      return true;
    }
    printStatus (240, "article posted ok") ;
    return true ;
  }
  
}
